from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pystac

from stac_generator.core.base import (
    CollectionGenerator,
    ItemGenerator,
    StacCollectionConfig,
)
from stac_generator.core.base.schema import SourceConfig
from stac_generator.core.base.utils import read_source_config
from stac_generator.core.point import PointGenerator
from stac_generator.core.point.schema import PointConfig
from stac_generator.core.raster import RasterGenerator
from stac_generator.core.raster.schema import RasterConfig
from stac_generator.core.vector import VectorGenerator
from stac_generator.core.vector.schema import VectorConfig

if TYPE_CHECKING:
    from concurrent.futures import Executor

    import pystac


EXTENSION_MAP: dict[str, type[SourceConfig]] = {
    "csv": PointConfig,
    "txt": PointConfig,
    "geotiff": RasterConfig,
    "tiff": RasterConfig,
    "tif": RasterConfig,
    "zip": VectorConfig,
    "geojson": VectorConfig,
    "json": VectorConfig,
    "gpkg": VectorConfig,  # Can also contain raster data. TODO: overhaul interface
    "shp": VectorConfig,
}

CONFIG_GENERATOR_MAP: dict[type[SourceConfig], type[ItemGenerator]] = {
    VectorConfig: VectorGenerator,
    RasterConfig: RasterGenerator,
    PointConfig: PointGenerator,
}

BaseConfig_T = (
    str
    | Path
    | SourceConfig
    | dict[str, Any]
    | Sequence[str]
    | Sequence[SourceConfig]
    | Sequence[dict[str, Any]]
)
Config_T = BaseConfig_T | Sequence[BaseConfig_T]


class StacGeneratorFactory:
    @staticmethod
    def get_extension_handler(extension: str) -> type[SourceConfig]:
        """Match file extension with SourceConfig type"""
        if extension not in EXTENSION_MAP:
            raise ValueError(
                f"No SourceConfig matches extension: {extension}. Either change the extension or register a handler with the method `register_extension_handler`"
            )
        return EXTENSION_MAP[extension]

    @staticmethod
    def register_extension_handler(
        extension: str, handler: type[SourceConfig], force: bool = False
    ) -> None:
        """Dynamically register extension handler"""
        if extension in EXTENSION_MAP and not force:
            raise ValueError(
                f"Handler for extension: {extension} already exists: {EXTENSION_MAP[extension].__name__}. If this is intentional, use register_extension_handler with force=True"
            )
        if not issubclass(handler, SourceConfig):
            raise ValueError("Registered handler must be an instance of a subclass of SourceConfig")
        EXTENSION_MAP[extension] = handler

    @staticmethod
    def register_generator_handler(
        config: SourceConfig,
        handler: type[ItemGenerator],
        force: bool = False,
    ) -> None:
        """Dynamically register a customer ItemGenerator class based on (new/existing) config type. Use force to overwrite existing handler"""
        config_type = type(config)
        if config_type in CONFIG_GENERATOR_MAP and not force:
            raise ValueError(
                f"Handler for config: {config_type.__name__} already exists: {CONFIG_GENERATOR_MAP[config_type].__name__}. If this is intentional, use register_generator_handler with force=True"
            )
        if not issubclass(handler, ItemGenerator):
            raise ValueError(
                "Registered handler must be an instance of a subclass of ItemGenerator"
            )
        CONFIG_GENERATOR_MAP[config_type] = handler

    @staticmethod
    def get_generator_handler(config: SourceConfig) -> type[ItemGenerator]:
        """Match ItemGenrator class based on config type"""
        config_type = type(config)
        if config_type not in CONFIG_GENERATOR_MAP:
            raise ValueError(
                f"No ItemGenerator for config of type: {config_type.__name__}. To register a handler, use the method StacGeneratorFactor.register_generator_handler"
            )
        return CONFIG_GENERATOR_MAP[config_type]

    @staticmethod
    def extract_item_config(item: pystac.Item) -> SourceConfig:
        """Get stac_generator properties. Used by the MCCN engine"""
        if "stac_generator" not in item.properties:
            raise ValueError(f"Missing stac_generator properties for item: {item.id}")
        ext = item.properties["stac_generator"]["location"].split(".")[-1]
        handler = StacGeneratorFactory.get_extension_handler(ext)
        return handler.model_validate(item.properties["stac_generator"])

    @staticmethod
    def get_item_generators(  # noqa: C901
        configs: Config_T,
    ) -> list[ItemGenerator]:
        def handle_dict_config(config_dict: dict[str, Any]) -> SourceConfig:
            if "id" not in config_dict:
                raise ValueError("Missing id in a config item.")
            if "location" not in config_dict:
                raise ValueError(f"Missing location in a config item: {config_dict['id']}")
            ext = config_dict["location"].split(".")[-1]
            config_handler = StacGeneratorFactory.get_extension_handler(ext)
            return config_handler(**config_dict)

        def handle_str_config(config_str: str) -> list[SourceConfig]:
            configs = read_source_config(config_str)
            return [handle_dict_config(config) for config in configs]

        def handle_base_config(
            config: str | dict[str, Any] | SourceConfig | Path,
        ) -> list[SourceConfig]:
            configs: list[SourceConfig] = []
            if isinstance(config, str):
                configs.extend(handle_str_config(config))
            elif isinstance(config, Path):
                configs.extend(handle_str_config(str(config)))
            elif isinstance(config, SourceConfig):
                configs.append(config)
            elif isinstance(config, dict):
                configs.append(handle_dict_config(config))
            else:
                raise TypeError(f"Invalid config item type: {type(config)}")
            return configs

        def handle_config(config: Config_T) -> list[SourceConfig]:
            if isinstance(config, str | dict | SourceConfig | Path):
                return handle_base_config(config)
            if hasattr(config, "__len__"):
                configs: list[SourceConfig] = []
                for item in config:
                    configs.extend(handle_config(item))
                return configs
            raise TypeError(f"Invalid config type: {type(config)}")

        parsed_configs = handle_config(configs)

        generators = []
        for config in parsed_configs:
            handler = StacGeneratorFactory.get_generator_handler(config)
            generators.append(handler(config))
        return generators

    @staticmethod
    def get_collection_generator(
        source_configs: Config_T,
        collection_config: StacCollectionConfig,
        pool: Executor | None = None,
    ) -> CollectionGenerator:
        """Get a CollectionGenerator instance based on source configs and
        collection config

        Args:
            source_configs (Config_T): extra metadata/generation parameters for
            the collection's items. Source configs can be given as:
            - a string path to a config.json file
            - a list of string paths to different config.json files
            - an instance of SourceConfig class (VectorConfig, PointConfig, RasterConfig), etc
            - a list of instances of SourceConfig.
            - a dictionary that follows the general structure of a SourceConfig class
            - a list of dictionaries
            If file paths are given, the files will be parsed into instances of SourceConfig.
            If dictionaries are given, they will be parsed into istances of SourceConfig using
            pydantic model_valiate.
            collection_config (StacCollectionConfig): collection metadata.
            pool (Executor | None): optional threadpool/process pool for parallel processing.

        Returns:
            CollectionGenerator: a collection generator instance, in which all items are derived
            from source _configs and general metadata derived from collection_config.
        """
        handlers = StacGeneratorFactory.get_item_generators(source_configs)
        return CollectionGenerator(collection_config, handlers, pool)



from stac_generator.core.base.generator import CollectionGenerator, ItemGenerator, StacSerialiser
from stac_generator.core.base.schema import SourceConfig, StacCollectionConfig

__all__ = (
    "CollectionGenerator",
    "ItemGenerator",
    "SourceConfig",
    "StacCollectionConfig",
    "StacSerialiser",
)



from typing import Any

EXCEPT_KEYS = None


def compare_dict_except(
    first: dict[Any, Any],
    second: dict[Any, Any],
    keys: set[str] | None = EXCEPT_KEYS,
) -> None:
    if not keys:
        assert first == second
    else:
        first_keys = set(first.keys()) - keys
        second_keys = set(second.keys()) - keys
        assert first_keys == second_keys
        for k in first_keys:
            assert first[k] == second[k]



from __future__ import annotations

import logging

import pystac

from stac_generator._types import CsvMediaType
from stac_generator.core.base.generator import VectorGenerator
from stac_generator.core.base.schema import ASSET_KEY
from stac_generator.core.base.utils import read_point_asset
from stac_generator.core.point.schema import PointConfig

logger = logging.getLogger(__name__)


class PointGenerator(VectorGenerator[PointConfig]):
    """ItemGenerator class that handles point data in csv format"""

    def generate(self) -> pystac.Item:
        """Create item from source csv config

        :param self.config: config which contains csv metadata
        :type self.config: PointConfig
        :return: stac metadata of the item described in self.config
        :rtype: pystac.Item
        """
        assets = {
            ASSET_KEY: pystac.Asset(
                href=self.config.location,
                description="Raw csv data",
                roles=["data"],
                media_type=CsvMediaType,
            )
        }
        raw_df = read_point_asset(
            self.config.location,
            self.config.X,
            self.config.Y,
            self.config.epsg,
            self.config.Z,
            self.config.T,
            self.config.date_format,
            self.config.column_info,
            self.config.timezone,
        )

        properties = self.config.to_properties()
        return self.df_to_item(
            raw_df,
            assets,
            self.config,
            properties={"stac_generator": properties},
            epsg=self.config.epsg,
            time_column=self.config.T,
        )



import json
from pathlib import Path

import pytest

from stac_generator.core.base.generator import CollectionGenerator
from stac_generator.core.base.schema import StacCollectionConfig
from stac_generator.core.base.utils import read_source_config
from stac_generator.core.raster.generator import RasterGenerator
from tests.utils import compare_dict_except

CONFIG_JSON = Path("tests/files/integration_tests/raster/config/raster_config.json")


GENERATED_DIR = Path("tests/files/integration_tests/raster/generated")


JSON_CONFIGS = read_source_config(str(CONFIG_JSON))
ITEM_IDS = [item["id"] for item in JSON_CONFIGS]


@pytest.fixture(scope="module")
def raster_generators() -> list[RasterGenerator]:
    return [RasterGenerator(config) for config in JSON_CONFIGS]


@pytest.fixture(scope="module")
def collection_generator(raster_generators: list[RasterGenerator]) -> CollectionGenerator:
    return CollectionGenerator(StacCollectionConfig(id="collection"), generators=raster_generators)


@pytest.mark.parametrize("item_idx", range(len(JSON_CONFIGS)), ids=ITEM_IDS)
def test_generator_given_item_expects_matched_generated_item(
    item_idx: int, raster_generators: list[RasterGenerator]
) -> None:
    config = JSON_CONFIGS[item_idx]
    expected_path = GENERATED_DIR / f"{config['id']}/{config['id']}.json"
    with expected_path.open() as file:
        expected = json.load(file)
    actual = raster_generators[item_idx].generate().to_dict()
    assert expected["id"] == actual["id"]
    assert expected["bbox"] == actual["bbox"]
    assert expected["geometry"] == actual["geometry"]
    compare_dict_except(expected["properties"], actual["properties"])
    assert expected["assets"] == actual["assets"]


def test_collection_generator(collection_generator: CollectionGenerator) -> None:
    actual = collection_generator().to_dict()
    expected_path = GENERATED_DIR / "collection.json"
    with expected_path.open() as file:
        expected = json.load(file)
    assert actual["extent"] == expected["extent"]



r"""
  /$$$$$$ /$$$$$$$$/$$$$$$  /$$$$$$         /$$$$$$ /$$$$$$$$/$$   /$$
 /$$__  $|__  $$__/$$__  $$/$$__  $$       /$$__  $| $$_____| $$$ | $$
| $$  \__/  | $$ | $$  \ $| $$  \__/      | $$  \__| $$     | $$$$| $$
|  $$$$$$   | $$ | $$$$$$$| $$            | $$ /$$$| $$$$$  | $$ $$ $$
 \____  $$  | $$ | $$__  $| $$            | $$|_  $| $$__/  | $$  $$$$
 /$$  \ $$  | $$ | $$  | $| $$    $$      | $$  \ $| $$     | $$\  $$$
|  $$$$$$/  | $$ | $$  | $|  $$$$$$/      |  $$$$$$| $$$$$$$| $$ \  $$
 \______/   |__/ |__/  |__/\______/        \______/|________|__/  \__/
"""  # noqa: D212

from __future__ import annotations

import json
import logging
from argparse import ArgumentParser, Namespace, _SubParsersAction
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path

from rich_argparse import RawDescriptionRichHelpFormatter

from stac_generator.__version__ import __version__

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def serialise_handler(args: Namespace) -> None:
    from stac_generator.core.base.generator import StacSerialiser
    from stac_generator.core.base.schema import StacCollectionConfig
    from stac_generator.factory import StacGeneratorFactory

    if args.v:
        import logging

        logging.getLogger().setLevel(logging.DEBUG)
    # Build collection config and catalog config
    metadata_json = {}
    if args.metadata_json:
        with Path(args.metadata_json).open("r") as file:
            metadata_json = json.load(file)

    # CLI args take precedence over metadata fields
    collection_config = StacCollectionConfig(
        id=args.id,
        title=args.title,
        description=args.description,
        license=args.license if args.license else metadata_json.get("license"),
        platform=args.platform if args.platform else metadata_json.get("platform"),
        constellation=args.constellation
        if args.constellation
        else metadata_json.get("constellation"),
        mission=args.mission if args.mission else metadata_json.get("mission"),
        instruments=args.instruments if args.instruments else metadata_json.get("instruments"),
        providers=metadata_json.get("providers"),
    )

    # Generate
    if args.num_workers == 1:
        # Use a single thread
        generator = StacGeneratorFactory.get_collection_generator(
            source_configs=args.src,
            collection_config=collection_config,
        )
        # Save
        serialiser = StacSerialiser(generator, args.dst)
        serialiser()
    elif args.num_workers > 1:
        with ProcessPoolExecutor(max_workers=args.num_workers) as executor:
            generator = StacGeneratorFactory.get_collection_generator(
                source_configs=args.src,
                collection_config=collection_config,
                pool=executor,
            )
            serialiser = StacSerialiser(generator, args.dst)
            serialiser()
    else:
        raise ValueError(f"Invalid number of threads: {args.num_workers}. Must be greater than 0.")


def add_serialise_sub_command(sub_parser: _SubParsersAction) -> None:
    parser = sub_parser.add_parser("serialise", help="Generate STAC record")
    # Source commands
    parser.add_argument(
        "src",
        type=str,
        action="extend",
        nargs="+",
        help="""path to the source_config.
                Path can be a local path or a url.
                Path also accepts multiple values.
                Source config contains metadata specifying how a raw file is read.
                At the minimum, it must contain the file location.
                To learn more about source config, please visit INSERT_DOC_URL.
            """,
    )
    parser.add_argument(
        "--dst",
        type=str,
        default="generated",
        help="""path to where the generated collection is stored.
                Accepts a local path or a remote api endpoint.
                If path is local, collection and item json files will be written to disk.
                If path is an endpoint, the collection and item json files will be pushed using STAC api methods.
                If not value is provided, the folder generated will be created in the current path to store generated records
            """,
    )

    parser.add_argument("-v", action="store_true", help="increase verbosity for debugging")
    # Collection Information
    collection_metadata = parser.add_argument_group("STAC collection metadata")
    collection_metadata.add_argument(
        "--id", type=str, help="id of collection", default="Collection"
    )
    collection_metadata.add_argument(
        "--title", type=str, help="title of collection", required=False, default="Auto-generated."
    )
    collection_metadata.add_argument(
        "--description",
        type=str,
        help="description of collection",
        required=False,
        default="Auto-generated",
    )

    # STAC Common Metadata
    common_metadata = parser.add_argument_group("STAC common metadata")
    common_metadata.add_argument(
        "--license", type=str, help="STAC license", required=False, default="proprietary"
    )
    common_metadata.add_argument(
        "--platform", type=str, help="STAC platform", required=False, default=None
    )
    common_metadata.add_argument(
        "--constellation", type=str, help="STAC constellation", required=False, default=None
    )
    common_metadata.add_argument(
        "--mission", type=str, help="STAC mission", required=False, default=None
    )
    common_metadata.add_argument(
        "--instruments",
        action="extend",
        nargs="+",
        type=str,
        required=False,
        default=None,
        help="STAC instrument",
    )
    common_metadata.add_argument(
        "--metadata_json",
        type=str,
        required=False,
        default=None,
        help="path to json file describing the metadata",
    )

    # Serialiser metadata
    serialiser_metadata = parser.add_argument_group("Serialliser metadata")
    serialiser_metadata.add_argument(
        "--num_workers",
        type=int,
        required=False,
        default=1,
        help="Number of threads to use for serialisation. If 1, serialisation will be done in a single thread.",
    )
    parser.set_defaults(func=serialise_handler)


def run_cli() -> None:
    # Build the CLI argument parser
    parser = ArgumentParser(
        prog="stac_generator",
        description=__doc__,
        formatter_class=RawDescriptionRichHelpFormatter,
    )
    parser.add_argument("-V", "--version", action="version", version=__version__)
    sub_parser = parser.add_subparsers(dest="command", help="Sub commands")
    add_serialise_sub_command(sub_parser)
    args = parser.parse_args()

    if args.command == "serialise":
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    run_cli()



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

import pandas as pd
import pystac

from stac_generator.core.base.generator import VectorGenerator as BaseVectorGenerator
from stac_generator.core.base.schema import ASSET_KEY
from stac_generator.core.base.utils import (
    get_timezone,
    read_join_asset,
    read_vector_asset,
)
from stac_generator.core.vector.schema import VectorConfig
from stac_generator.exceptions import StacConfigException

if TYPE_CHECKING:
    from pyproj.crs.crs import CRS

logger = logging.getLogger(__name__)


def extract_epsg(crs: CRS) -> tuple[int, bool]:
    """Extract epsg information from crs object.
    If epsg info can be extracted directly from crs, return that value.
    Otherwise, try to convert the crs info to WKT2 and extract EPSG using regex

    Note that this method may yield unreliable result

    :param crs: crs object
    :type crs: CRS
    :return: epsg information
    :rtype: tuple[int, bool] - epsg code and reliability flag
    """
    if (result := crs.to_epsg()) is not None:
        return (result, True)
    # Handle WKT1 edge case
    wkt = crs.to_wkt()
    match = re.search(r'ID\["EPSG",(\d+)\]', wkt)
    if match:
        return (int(match.group(1)), True)
    # No match - defaults to 4326
    logger.warning(
        "Cannot determine epsg from vector file. Either provide it in the config or change the source file. Defaults to 4326 but can be incorrect."
    )
    return (4326, False)


class VectorGenerator(BaseVectorGenerator[VectorConfig]):
    """ItemGenerator class that handles vector data with common vector formats - i.e (shp, zipped shp, gpkg, geojson)"""

    def generate(self) -> pystac.Item:
        """Create item from vector config

        :param self.config: config information
        :type self.config: VectorConfig
        :raises ValueError: if config epsg information is different from epsg information from vector file
        :return: stac metadata of the file described by self.config
        :rtype: pystac.Item
        """

        assets = {
            ASSET_KEY: pystac.Asset(
                href=str(self.config.location),
                media_type=pystac.MediaType.GEOJSON
                if self.config.location.endswith(".geojson")
                else "application/x-shapefile",
                roles=["data"],
                description="Raw vector data",
            )
        }
        logger.debug(f"Reading file from {self.config.location}")
        time_column = None
        # Only read relevant fields
        columns = [col["name"] if isinstance(col, dict) else col for col in self.config.column_info]
        # Throw exceptions if column_info contains invalid column
        raw_df = read_vector_asset(self.config.location, layer=self.config.layer)

        if columns and not set(columns).issubset(set(raw_df.columns)):
            raise StacConfigException(
                f"Invalid columns for asset - {self.config.location!s}: {set(columns) - set(raw_df.columns)}"
            )

        # Validate EPSG user-input vs extracted
        epsg, _ = extract_epsg(raw_df.crs)
        # Read join file
        if self.config.join_config:
            join_config = self.config.join_config
            # Get timezone information
            tzinfo = get_timezone(self.config.timezone, raw_df.to_crs(4326).geometry)
            # Try reading join file and raise errors if columns not provided
            join_df = read_join_asset(
                join_config.file,
                join_config.right_on,
                join_config.date_format,
                join_config.date_column,
                join_config.column_info,
                tzinfo,
            )
            # Try joining the files to validate results:
            raw_df = pd.merge(
                raw_df,
                join_df,
                left_on=join_config.left_on,
                right_on=join_config.right_on,
            )
            # Set asset start and end datetime based on date information
            if join_config.date_column:
                time_column = join_config.date_column

        # Make properties
        properties = self.config.to_properties()

        return self.df_to_item(
            raw_df,
            assets,
            self.config,
            properties={"stac_generator": properties},
            epsg=epsg,
            time_column=time_column,
        )



from __future__ import annotations

import abc
import datetime as pydatetime
import json
import logging
import pathlib
from typing import TYPE_CHECKING, Any, Generic, cast

import geopandas as gpd
import numpy as np
import pandas as pd
import pystac
from pyproj import CRS
from pystac.collection import Extent
from pystac.extensions.projection import ItemProjectionExtension
from pystac.utils import datetime_to_str
from shapely import (
    Geometry,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
    box,
    to_geojson,
)
from shapely.geometry import shape

from stac_generator.core.base.schema import (
    SourceConfig,
    StacCollectionConfig,
    T,
)
from stac_generator.core.base.utils import (
    add_timestamps,
    force_write_to_stac_api,
    get_timezone,
    href_is_stac_api_endpoint,
    localise_timezone,
    parse_href,
)

if TYPE_CHECKING:
    from collections.abc import Sequence
    from concurrent.futures import Executor

    from stac_generator._types import TimeSequence


logger = logging.getLogger(__name__)


def run_generator(generator: ItemGenerator) -> pystac.Item:
    return generator.generate()


class CollectionGenerator:
    """CollectionGenerator class. User should not need to subclass this class unless greater control over how collection is generated from items is needed."""

    def __init__(
        self,
        collection_config: StacCollectionConfig,
        generators: Sequence[ItemGenerator[T]],
        pool: Executor | None = None,
    ) -> None:
        """CollectionGenerator - generate collection from generators attribute

        :param collection_config: collection metadata
        :type collection_config: StacCollectionConfig
        :param generators: sequence of ItemGenerator subclasses.
        :type generators: Sequence[ItemGenerator]
        """
        self.collection_config = collection_config
        self.generators = generators
        self.pool = pool

    @staticmethod
    def spatial_extent(items: Sequence[pystac.Item]) -> pystac.SpatialExtent:
        """Extract a collection's spatial extent based on geometry information of its items

        :param items: a list of children items in the collection
        :type items: Sequence[pystac.Item]
        :return: a bbox enveloping all items
        :rtype: pystac.SpatialExtent
        """
        geometries: list[Geometry] = []
        for item in items:
            if (geo := item.geometry) is not None:
                geometries.append(shape(geo))
        geo_series = gpd.GeoSeries(data=geometries)
        bbox = geo_series.total_bounds.tolist()
        logger.debug(f"collection bbox: {bbox}")
        return pystac.SpatialExtent(bbox)

    @staticmethod
    def temporal_extent(items: Sequence[pystac.Item]) -> pystac.TemporalExtent:
        """Extract a collection's temporal extent based on time information of its items

        :param items: a list of children items in the collection
        :type items: Sequence[pystac.Item]
        :return: [start_time, end_time] enveloping all items
        :rtype: pystac.TemporalExtent
        """
        min_dt = pydatetime.datetime.now(pydatetime.UTC)
        max_dt = pydatetime.datetime(1, 1, 1, tzinfo=pydatetime.UTC)
        for item in items:
            if item.datetime is not None:
                min_dt = min(min_dt, item.datetime)
                max_dt = max(max_dt, item.datetime)
            else:
                raise ValueError(f"Unable to determine datetime for item: {item.id}")
        min_dt, max_dt = min([min_dt, max_dt]), max([max_dt, min_dt])
        logger.debug(
            f"collection time extent: {[datetime_to_str(min_dt), datetime_to_str(max_dt)]}"
        )
        return pystac.TemporalExtent([[min_dt, max_dt]])

    def _create_collection_from_items(
        self,
        items: Sequence[pystac.Item],
        collection_config: StacCollectionConfig | None = None,
    ) -> pystac.Collection:
        logger.debug("generating collection from items")
        if collection_config is None:
            raise ValueError("Generating collection requires non null collection config")
        collection = pystac.Collection(
            id=collection_config.id,
            description=(
                collection_config.description
                if collection_config.description
                else f"Auto-generated collection {collection_config.id} with stac_generator"
            ),
            extent=Extent(self.spatial_extent(items), self.temporal_extent(items)),
            title=collection_config.title,
            license=collection_config.license if collection_config.license else "proprietary",
            providers=[
                pystac.Provider.from_dict(item.model_dump()) for item in collection_config.providers
            ]
            if collection_config.providers
            else None,
        )
        collection.add_items(items)
        return collection

    def __call__(self) -> pystac.Collection:
        if self.pool:
            result = list(self.pool.map(run_generator, self.generators))
        else:
            result = [item.generate() for item in self.generators]
        return self._create_collection_from_items(result, self.collection_config)


class ItemGenerator(abc.ABC, Generic[T]):
    """Base ItemGenerator object. Users should extend this class for handling different file extensions."""

    source_type: type[T]
    """SourceConfig subclass that contains information used for parsing the source file"""

    @classmethod
    def __class_getitem__(cls, source_type: type) -> type:
        kwargs = {"source_type": source_type}
        return type(f"ItemGenerator[{source_type.__name__}]", (ItemGenerator,), kwargs)

    def __init__(
        self,
        config: dict[str, Any] | T,
    ) -> None:
        """Base ItemGenerator object. Users should extend this class for handling different file extensions.

        :param configs: source data configs - either from csv config or yaml/json
        :type configs: Mapping[str, Any]
        """
        logger.debug("validating config")
        if isinstance(config, self.source_type):
            self.config = config
        elif isinstance(config, dict):
            self.config = self.source_type(**config)
        else:
            raise TypeError(f"Invalid config type: {type(config)}")

    @abc.abstractmethod
    def generate(self) -> pystac.Item:
        """Abstract method that handles `pystac.Item` generation from the appropriate config"""
        raise NotImplementedError


class VectorGenerator(ItemGenerator[T]):
    @classmethod
    def __class_getitem__(cls, source_type: type) -> type:
        kwargs = {"source_type": source_type}
        return type(f"VectorGenerator[{source_type.__name__}]", (VectorGenerator,), kwargs)

    @staticmethod
    def geometry(  # noqa: C901
        df: gpd.GeoDataFrame,
    ) -> Geometry:
        """Calculate the geometry from geopandas dataframe.

        If geopandas dataframe has only one item, the geometry will be that of the item.
        If geopandas dataframe has less than 10 items of the same type, the geometry will be the Multi version of the type.
        Note that MultiPoint will be unpacked into points for the 10 items limit.
        If there are more than 10 items of the same type or there are items of different types i.e. Point and LineString, the returned
        geometry will be the Polygon of the bounding box. Note that Point and MultiPoint are treated as the same type (so are type and its Multi version).

        :param df: input dataframe
        :type df: gpd.GeoDataFrame
        """
        points: Sequence[Geometry] = df["geometry"].unique()
        # One item
        if len(points) == 1:
            return points[0]
        # Multiple Items of the same type
        curr_type = None
        curr_collection: list[Geometry] = []
        for point in points:
            if curr_type is None:
                match point:
                    case Point() | MultiPoint():
                        curr_type = MultiPoint
                    case LineString() | MultiLineString():
                        curr_type = MultiLineString
                    case Polygon() | MultiPolygon():
                        curr_type = MultiPolygon
                    case _:
                        return box(*df.total_bounds)
            if isinstance(point, Point) and curr_type == MultiPoint:
                curr_collection.append(point)
            elif isinstance(point, MultiPoint) and curr_type == MultiPoint:
                curr_collection.extend(point.geoms)
            elif isinstance(point, LineString) and curr_type == MultiLineString:
                curr_collection.append(point)
            elif isinstance(point, MultiLineString) and curr_type == MultiLineString:
                curr_collection.extend(point.geoms)
            elif isinstance(point, Polygon) and curr_type == MultiPolygon:
                curr_collection.append(point)
            elif isinstance(point, MultiPolygon) and curr_type == MultiPolygon:
                curr_collection.extend(point.geoms)
            else:
                return box(*df.total_bounds)
        if len(curr_collection) > 10:
            return box(*df.total_bounds)
        return cast(Geometry, curr_type)(curr_collection)

    @staticmethod
    def df_to_item(
        df: gpd.GeoDataFrame,
        assets: dict[str, pystac.Asset],
        source_config: SourceConfig,
        properties: dict[str, Any],
        epsg: int = 4326,
        time_column: str | None = None,
    ) -> pystac.Item:
        """Convert geopandas dataframe to pystac.Item

        :param df: input dataframe
        :type df: gpd.GeoDataFrame
        :param assets: source data asset_
        :type assets: dict[str, pystac.Asset]
        :param source_config: config
        :type source_config: SourceConfig
        :param properties: pystac Item properties
        :type properties: dict[str, Any]
        :param time_col: time_col if there are time information in the input df, defaults to None
        :type time_col: str | None, optional
        :param epsg: epsg information, defaults to 4326
        :type epsg: int, optional
        :return: generated pystac Item
        :rtype: pystac.Item
        """
        crs = cast(CRS, df.crs)
        # Convert to WGS 84 for computing geometry and bbox
        df.to_crs(epsg=4326, inplace=True)
        geometry = box(*df.total_bounds)
        item_tz = get_timezone(source_config.timezone, geometry)
        item_ts = source_config.get_datetime(geometry)

        geometry = json.loads(to_geojson(VectorGenerator.geometry(df)))

        # Process timestamps
        if time_column is None:
            # Item TS should be UTC by default
            timestamps: TimeSequence = [item_ts]
            start_datetime = item_ts
            end_datetime = item_ts
        else:
            sorted_ts = pd.Series(np.sort(df[time_column].unique()))  # Sorted unique values
            timestamps = localise_timezone(sorted_ts, item_tz)
            start_datetime = timestamps.min()
            end_datetime = timestamps.max()
        add_timestamps(properties, timestamps)

        item = pystac.Item(
            source_config.id,
            bbox=df.total_bounds.tolist(),
            geometry=geometry,
            datetime=item_ts,
            properties=properties,
            assets=assets,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        )
        proj_ext = ItemProjectionExtension.ext(item, add_if_missing=True)
        proj_ext.apply(epsg=epsg, wkt2=crs.to_wkt())
        return item


class StacSerialiser:
    def __init__(self, generator: CollectionGenerator, href: str) -> None:
        self.generator = generator
        self.collection = generator()
        self.href = href

    def pre_serialisation_hook(self, collection: pystac.Collection, href: str) -> None:
        """Hook that can be overwritten to provide pre-serialisation functionality.
        By default, this normalises collection href and performs validation

        :param collection: collection object
        :type collection: pystac.Collection
        :param href: serialisation href
        :type href: str
        """
        logger.debug("validating generated collection and items")
        collection.normalize_hrefs(href)
        collection.validate_all()

    def __call__(self) -> None:
        self.pre_serialisation_hook(self.collection, self.href)
        if href_is_stac_api_endpoint(self.href):
            self.to_json()
        else:
            self.to_api()
        logger.info(f"successfully save collection {self.collection.id} to {self.href}")

    @staticmethod
    def prepare_collection_configs(
        collection_generator: CollectionGenerator,
    ) -> list[dict[str, Any]]:
        items = collection_generator.generators
        result = []
        for item in items:
            result.append(StacSerialiser.prepare_config(item.config))
        return result

    @staticmethod
    def prepare_config(config: T) -> dict[str, Any]:
        return config.model_dump(
            mode="json",
            exclude_none=True,
            exclude_defaults=True,
            exclude_unset=True,
        )

    def save_collection_config(self, dst: str | pathlib.Path) -> None:
        config = self.prepare_collection_configs(self.generator)
        with pathlib.Path(dst).open("w") as file:
            json.dump(config, file)

    @staticmethod
    def save_configs(configs: Sequence[T], dst: str | pathlib.Path) -> None:
        config = [StacSerialiser.prepare_config(con) for con in configs]
        with pathlib.Path(dst).open("w") as file:
            json.dump(config, file)

    def to_json(self) -> None:
        """Generate STAC Collection and save to disk as json files"""
        logger.debug("saving collection as local json")
        self.collection.save()

    def to_api(self) -> None:
        """_Generate STAC Collection and push to remote API.
        The API will first attempt to send a POST request which will be replaced with a PUT request if a 409 error is encountered
        """
        logger.debug("save collection to STAC API")
        force_write_to_stac_api(
            url=parse_href(self.href, "collections"),
            id=self.collection.id,
            json=self.collection.to_dict(),
        )
        for item in self.collection.get_items(recursive=True):
            force_write_to_stac_api(
                url=parse_href(self.href, f"collections/{self.collection.id}/items"),
                id=item.id,
                json=item.to_dict(),
            )



from __future__ import annotations

import json
from typing import cast

import pystac
import rasterio
from pyproj import CRS
from pyproj.transformer import Transformer
from pystac.extensions.eo import Band, EOExtension
from pystac.extensions.projection import AssetProjectionExtension, ItemProjectionExtension
from pystac.extensions.raster import AssetRasterExtension, DataType, RasterBand
from shapely import box, to_geojson

from stac_generator.core.base.generator import ItemGenerator
from stac_generator.core.base.schema import ASSET_KEY
from stac_generator.core.base.utils import add_timestamps
from stac_generator.exceptions import SourceAssetException

from .schema import RasterConfig

VALID_COMMON_NAME = {
    "coastal",
    "blue",
    "green",
    "red",
    "rededge",
    "yellow",
    "pan",
    "nir",
    "nir08",
    "nir09",
    "cirrus",
    "swir16",
    "swir22",
    "lwir",
    "lwir11",
    "lwir12",
}


class RasterGenerator(ItemGenerator[RasterConfig]):
    """Raster Generator"""

    def generate(self) -> pystac.Item:
        """Generate Raster Item from config

        :param self.config: raster config - must contain band info
        :type self.config: RasterConfig
        :raises ValueError: if epsg code is provided in config but does not match that extracted from the asset
        :return: generated item
        :rtype: pystac.Item
        """
        try:
            with rasterio.open(self.config.location) as src:
                bounds = src.bounds
                crs = cast(CRS, src.crs)
                shape = list(src.shape)
        except rasterio.errors.RasterioIOError as e:
            raise SourceAssetException("Unable to read raster asset") from e

        # Convert to 4326 for bbox and geometry
        transformer = Transformer.from_crs(crs, 4326, always_xy=True)
        minx, miny = transformer.transform(bounds.left, bounds.bottom)
        maxx, maxy = transformer.transform(bounds.right, bounds.top)
        bbox: tuple[float, float, float, float] = (minx, miny, maxx, maxy)

        # Create geometry as Shapely Polygon
        geometry = box(*bbox)
        geometry_geojson = json.loads(to_geojson(geometry))

        # Process datetime
        item_ts = self.config.get_datetime(geometry)

        # Get EPSG
        epsg = crs.to_epsg()

        # Create STAC Item
        # Start datetime and end_datetime are set to be collection datetime for Raster data
        properties = {
            "stac_generator": self.config.to_properties(),
        }
        add_timestamps(properties, [item_ts])
        item = pystac.Item(
            id=self.config.id,
            geometry=geometry_geojson,
            bbox=list(bbox),
            datetime=item_ts,
            properties=properties,
            start_datetime=item_ts,
            end_datetime=item_ts,
        )

        # Projection extension
        proj_ext = ItemProjectionExtension.ext(item, add_if_missing=True)
        affine_transform = [
            rasterio.transform.from_bounds(*bounds, shape[1], shape[0])[i] for i in range(9)
        ]
        proj_ext.apply(epsg=epsg, wkt2=crs.to_wkt(), shape=shape, transform=affine_transform)

        # Create EO and Raster bands
        eo_bands = []
        raster_bands = []

        for band_info in self.config.band_info:
            common_name = band_info.get("common_name", None)
            if common_name and common_name not in VALID_COMMON_NAME:
                raise ValueError(f"Invalid common name: {common_name} for item: {self.config.id}")
            eo_band = Band.create(
                name=band_info["name"],
                common_name=common_name,
                center_wavelength=band_info.get("wavelength", None),
                description=band_info.get("description", None),
            )
            eo_bands.append(eo_band)

            raster_band = RasterBand.create(
                nodata=band_info.get("nodata", 0),
                data_type=DataType(band_info.get("data_type"))
                if band_info.get("data_type", None)
                else None,
            )
            raster_bands.append(raster_band)

        # Create Asset and Add to Item
        asset = pystac.Asset(
            href=self.config.location,
            media_type=pystac.MediaType.GEOTIFF,
            roles=["data"],
            title="Raster Data",
        )
        item.add_asset(ASSET_KEY, asset)

        # Apply projection extension to asset
        asset_proj_ext = AssetProjectionExtension.ext(asset, add_if_missing=True)
        asset_proj_ext.apply(
            epsg=crs.to_epsg(), wkt2=crs.to_wkt(), shape=shape, transform=affine_transform
        )

        # Apply Raster Extension to the Asset
        asset_raster_ext = AssetRasterExtension.ext(asset, add_if_missing=True)
        asset_raster_ext.apply(bands=raster_bands)

        # Add eo:bands to Asset
        # Apply EO bands to the asset using AssetEOExtension
        asset_eo_ext = EOExtension.ext(asset, add_if_missing=True)
        asset_eo_ext.apply(bands=eo_bands)

        return item



from stac_generator.core.base import (
    CollectionGenerator,
    ItemGenerator,
    SourceConfig,
    StacCollectionConfig,
    StacSerialiser,
)
from stac_generator.core.point import PointConfig, PointGenerator
from stac_generator.core.raster import RasterConfig, RasterGenerator
from stac_generator.core.vector import VectorConfig, VectorGenerator

__all__ = (
    "CollectionGenerator",
    "ItemGenerator",
    "PointConfig",
    "PointGenerator",
    "RasterConfig",
    "RasterGenerator",
    "SourceConfig",
    "StacCollectionConfig",
    "StacSerialiser",
    "VectorConfig",
    "VectorGenerator",
)



import json
from collections.abc import Generator
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import pytest

from stac_generator.core.base import StacCollectionConfig
from stac_generator.core.base.generator import CollectionGenerator
from stac_generator.factory import StacGeneratorFactory
from tests.utils import compare_dict_except

FILE_PATH = Path("tests/files/integration_tests")
GENERATED_PATH = FILE_PATH / "composite/generated"
COLLECTION_ID = "collection"
collection_config = StacCollectionConfig(id=COLLECTION_ID)
CONFIGS_LIST = [
    str(FILE_PATH / "point/config/point_config.csv"),
    str(FILE_PATH / "vector/config/vector_config.json"),
    str(FILE_PATH / "raster/config/raster_config.csv"),
]
COMPOSITE_CONFIG = FILE_PATH / "composite/config/composite_config.json"


@pytest.fixture(scope="module")
def composite_generator() -> CollectionGenerator:
    return StacGeneratorFactory.get_collection_generator(str(COMPOSITE_CONFIG), collection_config)


@pytest.fixture(scope="module")
def list_generator() -> CollectionGenerator:
    return StacGeneratorFactory.get_collection_generator(CONFIGS_LIST, collection_config)


@pytest.fixture(scope="module")
def threadpool_generator() -> Generator[CollectionGenerator, None, None]:
    executor = ThreadPoolExecutor(max_workers=4)
    # Use the executor to create a thread pool for the generator
    yield StacGeneratorFactory.get_collection_generator(
        CONFIGS_LIST,
        collection_config,
        pool=executor,
    )
    # Cleanup
    executor.shutdown(wait=True)


@pytest.mark.parametrize(
    "generator_fx",
    (
        "composite_generator",
        "list_generator",
        "threadpool_generator",
    ),
    ids=[
        "Composite Config",
        "List Configs",
        "ThreadPool Config",
    ],
)
def test_generator_factory(
    generator_fx: str,
    request: pytest.FixtureRequest,
) -> None:
    generator: CollectionGenerator = request.getfixturevalue(generator_fx)
    collection = generator()
    expected_collection_path = GENERATED_PATH / "collection.json"
    with expected_collection_path.open() as file:
        expected_collection = json.load(file)
    actual_collection = collection.to_dict()
    assert actual_collection["extent"] == expected_collection["extent"]
    for item in collection.get_items(recursive=True):
        config_loc = GENERATED_PATH / f"{item.id}/{item.id}.json"
        with config_loc.open("r") as file:
            expected = json.load(file)
        actual = item.to_dict()
        assert expected["id"] == actual["id"]
        assert expected["bbox"] == actual["bbox"]
        assert expected["geometry"] == actual["geometry"]
        compare_dict_except(expected["properties"], actual["properties"])
        assert expected["assets"] == actual["assets"]



from __future__ import annotations

import json
from typing import NotRequired

from pydantic import field_validator
from typing_extensions import TypedDict

from stac_generator.core.base.schema import SourceConfig


class BandInfo(TypedDict):
    """Band information for raster data"""

    name: str
    common_name: NotRequired[str]
    wavelength: NotRequired[float]
    nodata: NotRequired[float]
    data_type: NotRequired[str]
    description: NotRequired[str]


class RasterConfig(SourceConfig):
    """Configuration for raster data sources"""

    band_info: list[BandInfo]
    """List of band information - REQUIRED"""

    @field_validator("band_info", mode="before")
    @classmethod
    def parse_bands(cls, v: str | list) -> list[BandInfo]:
        if isinstance(v, list):
            return v
        if isinstance(v, str):
            parsed = json.loads(v)
            if not isinstance(parsed, list):
                raise ValueError("bands parameter expects a json serialisation of a lis of Band")
            return parsed
        raise ValueError(f"Invalid bands dtype: {type(v)}")

    @field_validator("band_info", mode="after")
    @classmethod
    def convert_lower(cls, v: list[BandInfo]) -> list[BandInfo]:
        for i in range(len(v)):
            v[i]["name"] = v[i]["name"].lower()
            if "common_name" in v[i]:
                v[i]["common_name"] = v[i]["common_name"].lower()
        return v



from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path

import pandas as pd
import pystac
import pytest

from stac_generator.core.base.utils import read_source_config
from stac_generator.core.vector.generator import VectorGenerator
from stac_generator.exceptions import StacConfigException

CONFIG_PATH = Path("tests/files/unit_tests/vectors/configs")


@lru_cache
def load_items(file: str) -> Sequence[pystac.Item]:
    config_path = CONFIG_PATH / file
    configs = read_source_config(str(config_path))
    generators = [VectorGenerator(config) for config in configs]
    return [generator.generate() for generator in generators]


@lru_cache
def load_item(file: str) -> pystac.Item:
    return load_items(file)[0]


def run_join_nodate_test(
    item: pystac.Item, location: str, start: str, end: str, timestamps: list[str]
) -> None:
    assert item.properties["stac_generator"]["column_info"] == [
        {"name": "Suburb_Name", "description": "Suburb_Name"}
    ]
    assert item.properties["stac_generator"]["join_config"]["column_info"] == [
        {"name": "Area", "description": "Area name"},
        {"name": "Distance", "description": "Driving Distance to CBD in km"},
        {
            "name": "Public_Transport",
            "description": "Time taken to reach CBD by public transport in minutes",
        },
        {"name": "Drive", "description": "Time taken to reach CBD by driving in minutes"},
        {"name": "Growth", "description": "Average 5 year growth in percentage in 2025"},
        {"name": "Yield", "description": "Average rental yield in 2025"},
    ]
    assert item.properties["stac_generator"]["join_config"]["file"] == location
    assert item.properties["stac_generator"]["join_config"]["right_on"] == "Area"
    assert item.properties["stac_generator"]["join_config"]["left_on"] == "Suburb_Name"
    assert item.properties["start_datetime"] == start
    assert item.properties["end_datetime"] == end
    assert item.properties["timestamps"] == timestamps


def run_join_with_date_test(
    item: pystac.Item, location: str, start: str, end: str, timestamps: list[str]
) -> None:
    assert "column_info" in item.properties["stac_generator"]
    assert item.properties["stac_generator"]["column_info"] == [
        {"name": "Suburb_Name", "description": "Suburb_Name"}
    ]
    assert "join_config" in item.properties["stac_generator"]
    assert item.properties["stac_generator"]["join_config"]["column_info"] == [
        {"name": "Area", "description": "Area Name"},
        {"name": "Sell_Price", "description": "Median Sales Price in 2025"},
        {"name": "Rent_Price", "description": "Median Rental Price in 2025"},
        {
            "name": "Sell/Rent",
            "description": "Ratio of Sales Price (in $1000) over Rental Price (in $)",
        },
    ]
    assert item.properties["stac_generator"]["join_config"]["file"] == location
    assert item.properties["stac_generator"]["join_config"]["right_on"] == "Area"
    assert item.properties["stac_generator"]["join_config"]["left_on"] == "Suburb_Name"
    assert item.properties["stac_generator"]["join_config"]["date_column"] == "Date"
    assert item.properties["start_datetime"] == start
    assert item.properties["end_datetime"] == end
    assert item.properties["timestamps"] == timestamps


def test_given_invalid_wrong_layer_expects_raises() -> None:
    with pytest.raises(Exception):
        load_items("invalid_wrong_layer.json")


def test_given_invalid_column_info_expects_raises() -> None:
    with pytest.raises(StacConfigException):
        load_items("invalid_column_info.json")


def test_given_no_column_info_expects_no_value_in_property() -> None:
    item = load_item("no_column_info.json")
    assert "column_info" not in item.properties["stac_generator"]


def test_given_column_info_expects_column_info_in_property() -> None:
    item = load_item("with_column_info.json")
    assert "column_info" in item.properties["stac_generator"]
    assert item.properties["stac_generator"]["column_info"] != []


def test_given_with_epsg_expects_epsg_in_property() -> None:
    item = load_item("with_epsg.json")
    assert "proj:code" in item.properties
    assert item.properties["proj:code"] == "EPSG:1168"


def test_given_layers_info_expects_multiple_layers() -> None:
    items = load_items("with_layer.json")
    assert len(items) == 2
    assert items[0].id == "Sunbury"
    assert items[1].id == "Werribee"


def test_given_join_file_invalid_config_left_on_undescribed_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_left_on_undescribed.json")


def test_given_join_file_invalid_no_join_column_info_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_no_join_column_info.json")


def test_given_join_file_invalid_config_no_left_on_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_no_left_on.json")


def test_given_join_file_invalid_config_no_right_on_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_no_right_on.json")


def test_given_join_file_invalid_config_right_on_undescribed_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_right_on_undescribed.json")


def test_given_join_file_invalid_wrong_join_column_info_expects_throw() -> None:
    with pytest.raises(StacConfigException):
        load_item("join_invalid_config_wrong_join_column_info.json")


def test_given_join_file_invalid_wrong_join_date_column_expects_throw() -> None:
    with pytest.raises(StacConfigException):
        load_item("join_invalid_config_wrong_join_date_column.json")


def test_given_join_file_invalid_wrong_left_on_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_wrong_left_on.json")


def test_given_join_file_invalid_wrong_right_on_expects_throw() -> None:
    with pytest.raises(ValueError):
        load_item("join_invalid_config_wrong_right_on.json")


def test_given_join_with_date_expects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date.json")
    run_join_with_date_test(
        item,
        "tests/files/unit_tests/vectors/price.csv",
        "2020-01-01T00:00:00Z",
        "2025-01-01T00:00:00Z",
        [
            "2020-01-01T00:00:00+00:00",
            "2024-01-01T00:00:00+00:00",
            "2025-01-01T00:00:00+00:00",
        ],
    )


def test_given_join_with_date_custom_tz_expects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date_custom_tz.json")
    # Custom timzone should not affect price since everything is in UTC
    run_join_with_date_test(
        item,
        "tests/files/unit_tests/vectors/price.csv",
        "2020-01-01T00:00:00Z",
        "2025-01-01T00:00:00Z",
        [
            "2020-01-01T00:00:00+00:00",
            "2024-01-01T00:00:00+00:00",
            "2025-01-01T00:00:00+00:00",
        ],
    )


def test_given_join_with_date_no_tzexpects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date_no_tz.json")
    run_join_with_date_test(
        item,
        "tests/files/unit_tests/vectors/price_no_tz.csv",
        "2020-01-01T00:00:00Z",
        "2025-01-01T00:00:00Z",
        [
            "2020-01-01T00:00:00+00:00",
            "2024-01-01T00:00:00+00:00",
            "2025-01-01T00:00:00+00:00",
        ],
    )


def test_given_join_with_date_no_tz_utc_timezone_expects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date_no_tz_utc_timezone.json")
    run_join_with_date_test(
        item,
        "tests/files/unit_tests/vectors/price_no_tz.csv",
        "2020-01-01T11:00:00Z",
        "2025-01-01T11:00:00Z",
        [
            "2020-01-01T11:00:00+00:00",
            "2024-01-01T11:00:00+00:00",
            "2025-01-01T11:00:00+00:00",
        ],
    )


def test_given_join_with_date_multi_tz_local_expects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date_multi_tz_local.json")
    run_join_with_date_test(
        item,
        "tests/files/unit_tests/vectors/price_multi_tz_multi_area.csv",
        "2020-01-01T00:00:00Z",
        "2025-01-07T00:00:00Z",
        [
            "2020-01-01T00:00:00+00:00",
            "2020-01-02T00:00:00+00:00",
            "2020-01-03T00:00:00+00:00",
            "2020-01-04T00:00:00+00:00",
            "2020-01-05T00:00:00+00:00",
            "2020-01-06T00:00:00+00:00",
            "2020-01-07T00:00:00+00:00",
            "2024-01-01T00:00:00+00:00",
            "2024-01-02T00:00:00+00:00",
            "2024-01-03T00:00:00+00:00",
            "2024-01-04T00:00:00+00:00",
            "2024-01-05T00:00:00+00:00",
            "2024-01-06T00:00:00+00:00",
            "2024-01-07T00:00:00+00:00",
            "2025-01-01T00:00:00+00:00",
            "2025-01-02T00:00:00+00:00",
            "2025-01-03T00:00:00+00:00",
            "2025-01-04T00:00:00+00:00",
            "2025-01-05T00:00:00+00:00",
            "2025-01-06T00:00:00+00:00",
            "2025-01-07T00:00:00+00:00",
        ],
    )


def test_given_join_with_date_multi_tz_sydney_expects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date_multi_tz_sydney.json")
    run_join_with_date_test(
        item,
        "tests/files/unit_tests/vectors/price_multi_tz_multi_area.csv",
        "2020-01-01T00:00:00Z",
        "2025-01-07T00:00:00Z",
        [
            "2020-01-01T00:00:00+00:00",
            "2020-01-02T00:00:00+00:00",
            "2020-01-03T00:00:00+00:00",
            "2020-01-04T00:00:00+00:00",
            "2020-01-05T00:00:00+00:00",
            "2020-01-06T00:00:00+00:00",
            "2020-01-07T00:00:00+00:00",
            "2024-01-01T00:00:00+00:00",
            "2024-01-02T00:00:00+00:00",
            "2024-01-03T00:00:00+00:00",
            "2024-01-04T00:00:00+00:00",
            "2024-01-05T00:00:00+00:00",
            "2024-01-06T00:00:00+00:00",
            "2024-01-07T00:00:00+00:00",
            "2025-01-01T00:00:00+00:00",
            "2025-01-02T00:00:00+00:00",
            "2025-01-03T00:00:00+00:00",
            "2025-01-04T00:00:00+00:00",
            "2025-01-05T00:00:00+00:00",
            "2025-01-06T00:00:00+00:00",
            "2025-01-07T00:00:00+00:00",
        ],
    )


# NOTE: this is a very special edge case
# When reading csv using pd.read_csv with parse_dates,
# if a date column has a mixture of tz-awared and non-tz-awared rows,
# the non-tz-awared will be populated with offsets of the nearest tz aware row
def test_given_join_with_date_multi_tz_utc_expects_correct_start_end_datetime() -> None:
    item = load_item("join_with_date_multi_tz_utc.json")
    run_join_with_date_test(
        item,
        "tests/files/unit_tests/vectors/price_multi_tz_multi_area.csv",
        "2020-01-01T00:00:00Z",
        "2025-01-07T00:00:00Z",
        [
            "2020-01-01T00:00:00+00:00",
            "2020-01-02T00:00:00+00:00",
            "2020-01-03T00:00:00+00:00",
            "2020-01-04T00:00:00+00:00",
            "2020-01-05T00:00:00+00:00",
            "2020-01-06T00:00:00+00:00",
            "2020-01-07T00:00:00+00:00",
            "2024-01-01T00:00:00+00:00",
            "2024-01-02T00:00:00+00:00",
            "2024-01-03T00:00:00+00:00",
            "2024-01-04T00:00:00+00:00",
            "2024-01-05T00:00:00+00:00",
            "2024-01-06T00:00:00+00:00",
            "2024-01-07T00:00:00+00:00",
            "2025-01-01T00:00:00+00:00",
            "2025-01-02T00:00:00+00:00",
            "2025-01-03T00:00:00+00:00",
            "2025-01-04T00:00:00+00:00",
            "2025-01-05T00:00:00+00:00",
            "2025-01-06T00:00:00+00:00",
            "2025-01-07T00:00:00+00:00",
        ],
    )


def test_given_join_no_date_expects_same_start_end_datetime() -> None:
    item = load_item("join_no_date.json")
    run_join_nodate_test(
        item,
        "tests/files/unit_tests/vectors/distance.csv",
        "2024-12-31T13:00:00Z",
        "2024-12-31T13:00:00Z",
        ["2024-12-31T13:00:00+00:00"],
    )
    assert pd.Timestamp(item.properties["start_datetime"]) == item.datetime


def test_given_join_no_date_adelaide_expects_same_start_end_datetime() -> None:
    item = load_item("join_no_date_adelaide.json")
    run_join_nodate_test(
        item,
        "tests/files/unit_tests/vectors/distance.csv",
        "2024-12-31T13:30:00Z",
        "2024-12-31T13:30:00Z",
        ["2024-12-31T13:30:00+00:00"],
    )
    assert pd.Timestamp(item.properties["start_datetime"]) == item.datetime


def test_given_join_no_date_melbourne_expects_same_start_end_datetime() -> None:
    item = load_item("join_no_date_melbourne.json")
    run_join_nodate_test(
        item,
        "tests/files/unit_tests/vectors/distance.csv",
        "2024-12-31T13:00:00Z",
        "2024-12-31T13:00:00Z",
        ["2024-12-31T13:00:00+00:00"],
    )
    assert pd.Timestamp(item.properties["start_datetime"]) == item.datetime


def test_given_join_no_date_utc_expects_same_start_end_datetime() -> None:
    item = load_item("join_no_date_utc.json")
    run_join_nodate_test(
        item,
        "tests/files/unit_tests/vectors/distance.csv",
        "2025-01-01T00:00:00Z",
        "2025-01-01T00:00:00Z",
        ["2025-01-01T00:00:00+00:00"],
    )
    assert pd.Timestamp(item.properties["start_datetime"]) == item.datetime



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from typing import Literal

import pandas as pd
from httpx._types import PrimitiveData

"""<a href="https://www.rfc-editor.org/rfc/rfc7111">Csv MIME type</a>"""
CsvMediaType = "text/csv"

"""<a href="https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types">Excel MIME type</a>"""
ExcelMediaType = "application/vnd.ms-excel"

"""HTTP Methods"""
HTTPMethod = Literal["GET", "POST", "PUT", "DELETE", "PATCH"]

"""HTTP URL Query Params"""
QueryParamTypes = (
    Mapping[str, PrimitiveData | Sequence[PrimitiveData]]
    | list[tuple[str, PrimitiveData]]
    | tuple[tuple[str, PrimitiveData], ...]
    | str
    | bytes
)

"""HTTP Header"""
HeaderTypes = (
    Mapping[str, str]
    | Mapping[bytes, bytes]
    | Sequence[tuple[str, str]]
    | Sequence[tuple[bytes, bytes]]
)

"""HTTP Cookies"""
CookieTypes = dict[str, str] | list[tuple[str, str]]

"""HTTP Body"""
RequestContent = str | bytes | Iterable[bytes]

"""STAC objects"""
StacEntityT = Literal["Item", "ItemCollection", "Collection", "Catalogue"]

"""STAC API HTTP Methods"""
StacAPIMethod = Literal["POST", "PUT"]

Timestamp = pd.Timestamp
TimeSeries = pd.Series
TimeSequence = Sequence[Timestamp] | TimeSeries



from __future__ import annotations

import json
import logging
import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, cast, overload

import geopandas as gpd
import httpx
import numpy as np
import pandas as pd
import pytz
import yaml
from pyogrio.errors import DataLayerError, DataSourceError
from shapely import Geometry, GeometryCollection, centroid
from timezonefinder import TimezoneFinder

from stac_generator.exceptions import (
    ConfigFormatException,
    InvalidExtensionException,
    SourceAssetException,
    SourceAssetLocationException,
    StacConfigException,
    TimezoneException,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from stac_generator._types import TimeSequence, TimeSeries, Timestamp
    from stac_generator.core.base.schema import ColumnInfo

SUPPORTED_URI_SCHEMES = ["http", "https"]
logger = logging.getLogger(__name__)

TZFinder = TimezoneFinder()


def parse_href(base_url: str, collection_id: str, item_id: str | None = None) -> str:
    """Generate href for collection or item based on id"""
    if item_id:
        return urllib.parse.urljoin(base_url, f"{collection_id}/{item_id}")
    return urllib.parse.urljoin(base_url, f"{collection_id}")


def href_is_stac_api_endpoint(href: str) -> bool:
    """Check if href points to a resource behind a stac api"""
    output = urllib.parse.urlsplit(href)
    return output.scheme == ""


def force_write_to_stac_api(url: str, id: str, json: dict[str, Any]) -> None:
    """Force write a json object to a stac api endpoint."""
    try:
        logger.debug(f"sending POST request to {url}")
        response = httpx.post(url=url, json=json)
        response.raise_for_status()
    except httpx.HTTPStatusError as err:
        if err.response.status_code == 409:
            logger.debug(f"sending PUT request to {url}")
            response = httpx.put(url=f"{url}/{id}", json=json)
            response.raise_for_status()
        else:
            raise err


def read_source_config(href: str) -> list[dict[str, Any]]:
    logger.debug(f"reading config file from {href}")
    if not href.endswith(("json", "yaml", "yml", "csv")):
        raise InvalidExtensionException(f"Expects one of json, yaml, yml, csv. Received: {href}")
    try:
        if href.endswith(".csv"):
            df = pd.read_csv(href)
            df.replace(np.nan, None, inplace=True)
            return cast(list[dict[str, Any]], df.to_dict("records"))
        if not href.startswith(("http", "https")):
            with Path(href).open("r") as file:
                if href.endswith(("yaml", "yml")):
                    result = yaml.safe_load(file)
                if href.endswith("json"):
                    result = json.load(file)
        else:
            response = httpx.get(href, follow_redirects=True)
            response.raise_for_status()
            if href.endswith("json"):
                result = response.json()
            if href.endswith(("yaml", "yml")):
                result = yaml.safe_load(response.content.decode("utf-8"))
    except Exception as e:
        raise StacConfigException(f"Unable to read config file from {href}") from e

    if isinstance(result, dict):
        return [result]
    if isinstance(result, list):
        return result
    raise ConfigFormatException(
        f"Expects config to be read as a list of dictionary. Provided: {type(result)}"
    )


def calculate_timezone(geometry: Geometry | Sequence[Geometry]) -> str:
    """Calculate timezone from geometry"""
    point = (
        centroid(geometry)
        if isinstance(geometry, Geometry)
        else centroid(GeometryCollection(list(geometry)))
    )
    # Use TimezoneFinder to get the timezone
    timezone_str = TZFinder.timezone_at(lng=point.x, lat=point.y)

    if not timezone_str:
        raise TimezoneException(
            f"Could not determine timezone for coordinates: lon={point.x}, lat={point.y}"
        )
    return timezone_str


def get_timezone(
    timezone: str | Literal["local", "utc"], geometry: Geometry | Sequence[Geometry]
) -> str:
    if timezone == "local":
        return calculate_timezone(geometry)
    return timezone


@overload
def localise_timezone(data: Timestamp, tzinfo: str) -> Timestamp: ...
@overload
def localise_timezone(data: TimeSeries, tzinfo: str) -> TimeSeries: ...


def localise_timezone(data: Timestamp | TimeSeries, tzinfo: str) -> Timestamp | TimeSeries:
    """Add timezone information to data then converts to UTC"""
    try:
        tz = pytz.timezone(tzinfo)
    except Exception as e:
        raise TimezoneException("Invalid timezone localisation") from e

    def localise(row: pd.Timestamp) -> pd.Timestamp:
        if row.tzinfo is None:
            row = row.tz_localize(tz)
        return row.tz_convert(pytz.timezone("UTC"))

    if isinstance(data, pd.Timestamp):
        return localise(data)
    return data.apply(localise)


def _read_csv(
    src_path: str,
    required: set[str] | Sequence[str] | None = None,
    optional: set[str] | Sequence[str] | None = None,
    date_col: str | None = None,
    date_format: str | None = "ISO8601",
    columns: set[str] | set[ColumnInfo] | Sequence[str] | Sequence[ColumnInfo] | None = None,
) -> pd.DataFrame:
    logger.debug(f"reading csv from path: {src_path}")
    parse_dates: list[str] | bool = [date_col] if isinstance(date_col, str) else False
    usecols: set[str] | None = None
    # If band info is provided, only read in the required columns + the X and Y coordinates
    if columns:
        usecols = {item["name"] if isinstance(item, dict) else item for item in columns}
        if required:
            usecols.update(required)
        if optional:
            usecols.update(optional)
        if date_col:
            usecols.add(date_col)
    try:
        return pd.read_csv(
            filepath_or_buffer=src_path,
            usecols=list(usecols) if usecols else None,
            date_format=date_format,
            parse_dates=parse_dates,
        )
    except FileNotFoundError as e:
        raise SourceAssetLocationException(e) from None
    except ValueError as e:
        raise StacConfigException(
            f"Unable to read {src_path} using additional configuration parameters"
        ) from e


def read_point_asset(
    src_path: str,
    X_coord: str,
    Y_coord: str,
    epsg: int,
    Z_coord: str | None = None,
    T_coord: str | None = None,
    date_format: str = "ISO8601",
    columns: set[str] | set[ColumnInfo] | Sequence[str] | Sequence[ColumnInfo] | None = None,
    timezone: str | Literal["utc", "local"] = "local",
) -> gpd.GeoDataFrame:
    """Read in csv from local disk
    Users must provide at the bare minimum the location of the csv, and the names of the columns to be
    treated as the X and Y coordinates. By default, will read in all columns in the csv. If columns and groupby
    columns are provided, will selectively read specified columns together with the coordinate columns (X, Y, T).
    """
    df = _read_csv(
        src_path=src_path,
        required=[X_coord, Y_coord],
        optional=[Z_coord] if Z_coord else None,
        date_col=T_coord,
        date_format=date_format,
        columns=columns,
    )

    gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(df[X_coord], df[Y_coord], crs=epsg))
    if T_coord:
        tzinfo = get_timezone(timezone, gdf.geometry)
        gdf[T_coord] = localise_timezone(gdf[T_coord], tzinfo)
    return gdf


def read_vector_asset(
    src_path: str | Path,
    bbox: tuple[float, float, float, float] | None = None,
    columns: set[str] | Sequence[str] | None = None,
    layer: str | int | None = None,
) -> gpd.GeoDataFrame:
    try:
        return gpd.read_file(
            filename=src_path,
            bbox=bbox,
            columns=columns,
            layer=layer,
            engine="pyogrio",  # For predictability
        )
    except DataLayerError:
        raise StacConfigException(f"Invalid layer. File: {src_path}, layer: {layer}") from None
    except DataSourceError as e:
        raise SourceAssetException(e) from None


def read_join_asset(
    src_path: str,
    right_on: str,
    date_format: str,
    date_column: str | None,
    columns: set[str] | Sequence[str] | set[ColumnInfo] | Sequence[ColumnInfo],
    tzinfo: str,
) -> pd.DataFrame:
    df = _read_csv(
        src_path=src_path,
        required=[right_on],
        date_format=date_format,
        date_col=date_column,
        columns=columns,
    )
    if date_column:
        df[date_column] = localise_timezone(df[date_column], tzinfo)
    return df


def add_timestamps(properties: dict[Any, Any], timestamps: TimeSequence) -> None:
    timestamps_str = [item.isoformat(sep="T") for item in timestamps]
    properties["timestamps"] = timestamps_str



import datetime
import json
import logging
from collections.abc import Sequence
from typing import Any, Literal, NotRequired, Required, TypeVar

import pandas as pd
import pytz
from httpx._types import RequestData
from pydantic import BaseModel, Field, field_validator
from shapely import Geometry
from stac_pydantic.shared import Provider
from typing_extensions import TypedDict

from stac_generator._types import (
    CookieTypes,
    HeaderTypes,
    HTTPMethod,
    QueryParamTypes,
    RequestContent,
)
from stac_generator.core.base.utils import get_timezone
from stac_generator.exceptions import TimezoneException

T = TypeVar("T", bound="SourceConfig")
ASSET_KEY = "data"
logger = logging.getLogger(__name__)


class StacCollectionConfig(BaseModel):
    """Contains parameters to pass to Collection constructor. Also contains other metadata except for datetime related metadata.

    Collection's datetime, start_datetime and end_datetime will be derived from the time information of its children items

    This config provides additional information that can not be derived from source file, which includes
    <a href="https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md">Stac Common Metadata</a>
    and other descriptive information such as the id of the new entity
    """

    # Stac Information
    id: str
    """Item id"""
    title: str | None = "Auto-generated Stac Item"
    """A human readable title describing the item entity. https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#basics"""
    description: str | None = "Auto-generated Stac Item"
    """Detailed multi-line description to fully explain the STAC entity. https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#basics"""
    license: str | None = None
    """License(s) of the data as SPDX License identifier, SPDX License expression, or other - https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#licensing"""
    providers: list[Provider] | None = None
    """A list of providers, which may include all organizations capturing or processing the data or the hosting provider. Providers should be listed in chronological order with the most recent provider being the last element of the list. - https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#provider"""
    platform: str | None = None
    """Unique name of the specific platform to which the instrument is attached. https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#platform"""
    instruments: list[str] | None = None
    """Name of instrument or sensor used (e.g., MODIS, ASTER, OLI, Canon F-1). https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#instrument"""
    constellation: str | None = None
    """Name of the constellation to which the platform belongs. https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#constellation"""
    mission: str | None = None
    """Name of the mission for which data is collected. https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md#mission"""


class StacItemConfig(StacCollectionConfig):
    """Contains parameters to pass to Item constructor. Also contains other metadata except for datetime related metadata.

    Item's datetime will be superseded by `collection_date` and `collection_time` recorded in local timezone. The STAC `datetime`
    metadata is obtained from the method `get_datetime` by providing the local timezone, which will be automatically derived from
    the crs information.

    This config provides additional information that can not be derived from source file, which includes
    <a href="https://github.com/radiantearth/stac-spec/blob/master/commons/common-metadata.md">Stac Common Metadata</a>
    and other descriptive information such as the id of the new entity
    """

    collection_date: datetime.date
    """Date when the data is collected"""
    collection_time: datetime.time
    """Time when the data is collected"""
    timezone: str | Literal["utc", "local"] = "local"
    """Timezone"""

    def get_datetime(self, geometry: Geometry | Sequence[Geometry]) -> pd.Timestamp:
        timezone = get_timezone(self.timezone, geometry)
        try:
            local_dt = pd.Timestamp(
                f"{self.collection_date}T{self.collection_time}",
                tzinfo=pytz.timezone(timezone),
            )
        except Exception as e:
            raise TimezoneException("Invalid timezone config parameter") from e
        return local_dt.tz_convert(pytz.timezone("UTC"))


class SourceConfig(StacItemConfig):
    """Base source config that should be subclassed for different file extensions.

    Source files contain raw spatial information (i.e. geotiff, shp, csv) from which
    some Stac metadata can be derived. SourceConfig describes:

    - The access mechanisms for the source file: stored on local disk, or hosted somewhere behind an api endpoint. If the source
    file must be accessed through an endpoint, users can provide additional HTTP information that forms the HTTP request to the host server.
    - Processing information that are unique for the source type. Users should inherit `SourceConfig` for file extensions
    currently unsupported.
    - Additional Stac Metadata from `StacConfig`
    """

    location: str
    """Asset's href"""
    extension: str | None = None
    """Explicit file extension specification. If the file is stored behind an api endpoint, the field `extension` must be provided"""
    # HTTP Parameters
    method: HTTPMethod | None = "GET"
    """HTTPMethod to acquire the file from `location`"""
    params: QueryParamTypes | None = None
    """HTTP query params for getting file from `location`"""
    headers: HeaderTypes | None = None
    """HTTP query headers for getting file from `location`"""
    cookies: CookieTypes | None = None
    """HTTP query cookies for getting file from `location`"""
    content: RequestContent | None = None
    """HTTP query body content for getting file from `location`"""
    data: RequestData | None = None
    """HTTP query body content for getting file from `location`"""
    json_body: Any = None
    """HTTP query body content for getting file from `location`"""

    def to_properties(self) -> dict[str, Any]:
        return self.model_dump(mode="json", exclude_unset=True, exclude_none=True)


DTYPE = Literal[
    "str",
    "int",
    "bool",
    "float",
    "int8",
    "int16",
    "int32",
    "int64",
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "float16",
    "float32",
    "float64",
    "cint16",
    "cint32",
    "cfloat32",
    "cfloat64",
    "other",
]


class ColumnInfo(TypedDict):
    """TypedDict description of GeoDataFrame columns. Used for describing vector/point attributes"""

    name: Required[str]
    """Column name"""
    description: NotRequired[str]
    """Column description"""
    dtype: NotRequired[DTYPE]
    """Column data type"""


class HasColumnInfo(BaseModel):
    column_info: list[ColumnInfo] = Field(default_factory=list)
    """List of attributes associated with point/vector data"""

    @field_validator("column_info", mode="before")
    @classmethod
    def coerce_to_object(cls, v: str | list[ColumnInfo] | None) -> list[ColumnInfo]:
        """Convert json serialised string of column info into matched object"""
        if v is None:
            return []
        if isinstance(v, list):
            return v
        parsed = json.loads(v)
        if not isinstance(parsed, list):
            raise ValueError(
                "column_info field expects a json serialisation of a list of ColumnInfo or a list of string"
            )
        return parsed



from functools import lru_cache
from pathlib import Path

import pandas as pd
import pystac
import pytest

from stac_generator.core.base.utils import read_source_config
from stac_generator.core.point.generator import PointGenerator
from stac_generator.exceptions import StacConfigException, TimezoneException

CONFIG_PATH = Path("tests/files/unit_tests/points/configs")


@lru_cache
def load_item(file: str) -> pystac.Item:
    config_path = CONFIG_PATH / file
    config = read_source_config(str(config_path))
    generator = PointGenerator(config[0])
    return generator.generate()


def test_no_date_expects_start_datetime_end_datetime_same_as_datetime() -> None:
    item = load_item("no_date.json")
    assert item.datetime == pd.Timestamp(item.properties["start_datetime"])
    assert item.datetime == pd.Timestamp(item.properties["end_datetime"])
    assert item.properties["timestamps"] == [pd.Timestamp(item.datetime).isoformat()]


def test_non_default_fields_expects_same_properties() -> None:
    item = load_item("non_default_fields.json")
    assert item.properties["stac_generator"]["description"] == "Non default description"
    assert item.properties["stac_generator"]["license"] == "MIT"


def test_with_altitude_expects_Z_value_in_properties() -> None:
    item = load_item("with_altitude.json")
    assert item.properties["stac_generator"]["Z"] == "elevation"


def test_with_column_info_expects_column_info_in_properties() -> None:
    item = load_item("with_column_info.json")
    assert "column_info" in item.properties["stac_generator"]


def test_no_column_info_expects_no_column_info_in_properties() -> None:
    item = load_item("no_column_info.json")
    assert "column_info" not in item.properties["stac_generator"]


def test_with_date_no_tzinfo_expects_utc_start_end_datetime() -> None:
    item = load_item("with_date_no_tzinfo.json")
    assert item.properties["start_datetime"] == "2022-12-31T13:30:00Z"
    assert item.properties["end_datetime"] == "2023-01-02T13:30:00Z"
    date_range = pd.date_range("2022-12-31T13:30:00Z", "2023-01-02T13:30:00Z")
    exp_ts = [item.isoformat() for item in date_range]
    assert item.properties["timestamps"] == exp_ts


def test_with_date_with_utc_tz_expects_utc_start_end_datetime() -> None:
    item = load_item("with_date_with_utc_tz.json")
    assert item.properties["start_datetime"] == "2023-01-01T00:00:00Z"
    assert item.properties["end_datetime"] == "2023-01-03T00:00:00Z"
    date_range = pd.date_range("2023-01-01T00:00:00Z", "2023-01-03T00:00:00Z")
    exp_ts = [item.isoformat() for item in date_range]
    assert item.properties["timestamps"] == exp_ts


def test_no_date_with_utc_tz_expects_utc_start_end_datetime() -> None:
    item = load_item("no_date_with_utc_tz.json")
    assert item.datetime == pd.Timestamp(item.properties["start_datetime"])
    assert item.properties["start_datetime"] == "2017-01-01T00:00:00Z"
    assert item.properties["end_datetime"] == item.properties["start_datetime"]
    assert item.properties["timestamps"] == [pd.Timestamp(item.datetime).isoformat()]


def test_with_date_with_tzinfo_expects_start_end_datetime() -> None:
    item = load_item("with_date_with_tzinfo.json")
    assert item.properties["start_datetime"] == "2023-01-01T00:00:00Z"
    assert item.properties["end_datetime"] == "2023-01-03T00:00:00Z"
    date_range = pd.date_range("2023-01-01T00:00:00Z", "2023-01-03T00:00:00Z")
    exp_ts = [item.isoformat() for item in date_range]
    assert item.properties["timestamps"] == exp_ts


def test_with_date_with_date_multi_tz_unsorted_utc_expects_start_end_datetime() -> None:
    item = load_item("with_date_multi_tz_unsorted_utc.json")
    assert item.properties["start_datetime"] == "2023-01-01T00:00:00Z"
    assert item.properties["end_datetime"] == "2023-01-03T00:00:00Z"
    date_range = pd.date_range("2023-01-01T00:00:00Z", "2023-01-03T00:00:00Z")
    exp_ts = [item.isoformat() for item in date_range]
    assert item.properties["timestamps"] == exp_ts


def test_with_date_with_date_multi_tz_unsorted_local_expects_start_end_datetime() -> None:
    item = load_item("with_date_multi_tz_unsorted_local.json")
    assert item.properties["start_datetime"] == "2022-12-31T13:30:00Z"
    assert item.properties["end_datetime"] == "2023-01-03T00:00:00Z"
    exp_ts = [
        "2022-12-31T13:30:00+00:00",
        "2023-01-01T00:00:00+00:00",
        "2023-01-01T13:30:00+00:00",
        "2023-01-02T00:00:00+00:00",
        "2023-01-02T13:30:00+00:00",
        "2023-01-03T00:00:00+00:00",
    ]
    assert item.properties["timestamps"] == exp_ts


def test_invalid_altitude_expects_raises() -> None:
    with pytest.raises(StacConfigException):
        load_item("invalid_altitude.json")


def test_invalid_timezone_expects_raises() -> None:
    with pytest.raises(TimezoneException):
        load_item("invalid_tz.json")


def test_invalid_date_expects_raises() -> None:
    with pytest.raises(StacConfigException):
        load_item("invalid_date.json")


def test_invalid_column_info_expects_raises() -> None:
    with pytest.raises(StacConfigException):
        load_item("invalid_column_info.json")



from __future__ import annotations

from typing import Self

from pydantic import BaseModel, field_validator, model_validator

from stac_generator.core.base.schema import ColumnInfo, HasColumnInfo, SourceConfig


class JoinConfig(BaseModel):
    file: str
    left_on: str
    right_on: str
    date_column: str | None = None
    date_format: str = "ISO8601"
    column_info: list[ColumnInfo]

    @field_validator("column_info", mode="after")
    @classmethod
    def check_non_empty_column_info(cls, value: list[ColumnInfo]) -> list[ColumnInfo]:
        if not value:
            raise ValueError("Join file must have non-empty column_info")
        return value


class VectorConfig(SourceConfig, HasColumnInfo):
    """Extended source config with EPSG code."""

    layer: str | None = None
    """Vector layer for multi-layer shapefile"""

    join_config: JoinConfig | None = None
    """Config for join asset if valid"""

    @model_validator(mode="after")
    def check_join_fields_described(self) -> Self:
        if self.join_config:
            vector_columns = {col["name"] for col in self.column_info}
            join_columns = {col["name"] for col in self.join_config.column_info}
            if self.join_config.left_on not in vector_columns:
                raise ValueError("Join field must be described using column_info")
            if self.join_config.right_on not in join_columns:
                raise ValueError("Join field must be described using join file column_info")
        return self



import datetime as pydatetime

import geopandas as gpd
import pystac
import pytest
from shapely import Geometry

from stac_generator._types import CsvMediaType
from stac_generator.core.base.generator import VectorGenerator
from stac_generator.core.base.schema import ASSET_KEY
from stac_generator.core.base.utils import read_point_asset
from stac_generator.core.point.schema import PointConfig

ALL_COLUMNS = {
    "latitude",
    "longitude",
    "elevation",
    "station",
    "YYYY-MM-DD",
    "daily_rain",
    "max_temp",
    "min_temp",
    "radiation",
    "mslp",
}
X = "longitude"
Y = "latitude"
Z = "elevation"
T = "YYYY-MM-DD"
EPSG = 7843  # GDA2020
DATE_FORMAT = "ISO8601"
COLLECTION_DATE = pydatetime.date(2011, 1, 1)
COLLECTION_TIME = pydatetime.time(12, 4, 5)

PATHS = {
    "with_date_multi": "tests/files/unit_tests/points/with_date_multi.csv",
    "with_date_one": "tests/files/unit_tests/points/with_date_one.csv",
    "no_date_multi": "tests/files/unit_tests/points/no_date_multi.csv",
    "no_date_one": "tests/files/unit_tests/points/no_date_one.csv",
}
FRAMES = {
    "with_date_multi": read_point_asset(
        "tests/files/unit_tests/points/with_date_multi.csv", X, Y, EPSG, Z, T, DATE_FORMAT
    ),
    "with_date_one": read_point_asset(
        "tests/files/unit_tests/points/with_date_one.csv", X, Y, EPSG, Z, T, DATE_FORMAT
    ),
    "no_date_multi": read_point_asset(
        "tests/files/unit_tests/points/no_date_multi.csv", X, Y, EPSG
    ),
    "no_date_one": read_point_asset("tests/files/unit_tests/points/no_date_one.csv", X, Y, EPSG),
}
ASSETS = {
    key: pystac.Asset(href=value, roles=["data"], media_type=CsvMediaType)
    for key, value in PATHS.items()
}
CONFIGS = {
    "with_date_multi": PointConfig(
        X=X,
        Y=Y,
        T=T,
        id="test_id",
        location=PATHS["with_date_multi"],
        collection_date=COLLECTION_DATE,
        collection_time=COLLECTION_TIME,
    ),
    "with_date_one": PointConfig(
        X=X,
        Y=Y,
        T=T,
        id="test_id",
        location=PATHS["with_date_one"],
        collection_date=COLLECTION_DATE,
        collection_time=COLLECTION_TIME,
    ),
    "no_date_multi": PointConfig(
        X=X,
        Y=Y,
        id="test_id",
        location=PATHS["no_date_multi"],
        collection_date=COLLECTION_DATE,
        collection_time=COLLECTION_TIME,
    ),
    "no_date_one": PointConfig(
        X=X,
        Y=Y,
        id="test_id",
        location=PATHS["no_date_one"],
        collection_date=COLLECTION_DATE,
        collection_time=COLLECTION_TIME,
    ),
}
GEOMETRIES = {
    "with_date_multi": {
        "type": "MultiPoint",
        "coordinates": [
            [138.5196, -34.9524],
            [138.5296, -34.9624],
            [138.5396, -34.9724],
            [138.5496, -34.9824],
        ],
    },
    "with_date_one": {"type": "Point", "coordinates": [138.5196, -34.9524]},
    "no_date_multi": {
        "type": "MultiPoint",
        "coordinates": [[150.5505183, -24.34031206], [149.8055563, -29.04132741]],
    },
    "no_date_one": {"type": "Point", "coordinates": [150.3125397, -28.18249244]},
}


def test_read_point_asset_given_no_args_read_all_columns() -> None:
    df = read_point_asset(PATHS["with_date_one"], X, Y, epsg=EPSG)
    expected = set(ALL_COLUMNS) | {"geometry"}
    assert set(df.columns) == expected


@pytest.mark.parametrize(
    "z_col, t_col, columns",
    [
        (Z, T, ["max_temp", "min_temp"]),
        (Z, None, ["max_temp", "min_temp"]),
        (None, T, ["max_temp", "min_temp"]),
        (None, None, ["max_temp"]),
    ],
)
def test_read_point_asset_given_selected_columns_read_selected_columns(
    z_col: str | None,
    t_col: str | None,
    columns: list[str],
) -> None:
    df = read_point_asset(
        PATHS["with_date_one"],
        X,
        Y,
        epsg=EPSG,
        Z_coord=z_col,
        T_coord=t_col,
        columns=columns,
    )
    expected_columns = {X, Y, "geometry"}
    if z_col is not None:
        expected_columns.add(z_col)
    if t_col is not None:
        expected_columns.add(t_col)
    expected_columns = expected_columns | set(columns)
    assert set(df.columns) == expected_columns


@pytest.mark.parametrize(
    "frame, asset, source_config, geometry",
    zip(FRAMES.values(), ASSETS.values(), CONFIGS.values(), GEOMETRIES.values()),
    ids=FRAMES.keys(),
)
def test_df_to_item(
    frame: gpd.GeoDataFrame,
    asset: pystac.Asset,
    source_config: PointConfig,
    geometry: Geometry,
) -> None:
    item = VectorGenerator.df_to_item(
        df=frame,
        assets={ASSET_KEY: asset},
        source_config=source_config,
        properties={},
        epsg=source_config.epsg,
    )
    assert item.id == source_config.id
    assert item.datetime is not None
    assert item.assets == {ASSET_KEY: asset}
    assert item.geometry == geometry
    assert "proj:code" in item.properties
    assert "proj:wkt2" in item.properties



class SourceAssetException(Exception):
    """Exception raised when the source data cannot be read by itself"""


class SourceAssetLocationException(SourceAssetException):
    """Exception raised when the source asset location cannot be accessed"""


class TimezoneException(Exception):
    """Exception raised when timezone information cannot be determined from geometry"""


class StacConfigException(Exception):
    """Exception raised when the config metadata is misspecified"""


class ConfigFormatException(StacConfigException):
    """Exception raised when the format of the config is invalid"""


class InvalidExtensionException(StacConfigException):
    """Exception raised when the config extension is unsupported"""


class ColumnInfoException(StacConfigException):
    """Exception raised when information described in column info does not match source asset"""


class BandInfoException(StacConfigException):
    """Exception raised when information described in band info does not match raster asset"""


class JoinConfigException(StacConfigException):
    """Exception raised when information provided in join config is invalid"""



import datetime
import json

import geopandas as gpd
import httpx
import pandas as pd
import pystac
import pytest
import pytest_httpx
import shapely
from shapely import Geometry, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon

from stac_generator.core.base.generator import CollectionGenerator, VectorGenerator
from stac_generator.core.base.utils import (
    _read_csv,
    force_write_to_stac_api,
    href_is_stac_api_endpoint,
    localise_timezone,
    parse_href,
    read_source_config,
    read_vector_asset,
)
from stac_generator.exceptions import SourceAssetException, StacConfigException, TimezoneException

VALID_CSV_CONFIG_FILE = "tests/files/unit_tests/configs/csv_config.csv"

VALID_CONFIG_FILES = [
    "tests/files/unit_tests/configs/valid_json_config.json",
    "tests/files/unit_tests/configs/valid_yaml_config.yaml",
    "tests/files/unit_tests/configs/valid_yml_config.yml",
    "tests/files/unit_tests/configs/valid_json_config_one_item.json",
]


CONFIG_OUTPUT = [
    {
        "id": "Grandview",
        "location": "example/csv/Grandview.csv",
        "X": "Longitude",
        "Y": "Latitude",
        "epsg": 4326,
        "column_info": [
            {"name": "Ca_Soln", "description": "Ca concentration"},
            {"name": "pH", "description": "soil pH"},
        ],
        "collection_date": "2017-01-01",
        "collection_time": "00:00:00",
    }
]

INVALID_CONFIG_FILES = [
    # Empty files
    "tests/files/unit_tests/configs/invalid_json_config.json",
    "tests/files/unit_tests/configs/invalid_yaml_config.yaml",
    "tests/files/unit_tests/configs/invalid_yml_config.yml",
    # Invalid file extensions
    "tests/files/unit_tests/configs/invalid_extension_config.ini",
    "tests/files/unit_tests/configs/invalid_extension_config.toml",
    "tests/files/unit_tests/configs/invalid_extension_config.config",
    "tests/files/unit_tests/configs/invalid_extension_config.txt",
]

INVALID_CONFIG_FILES_IDS = [
    "empty_json",
    "empty_yaml",
    "empty_yml",
    "invalid_extension_ini",
    "invalid_extension_toml",
    "invalid_extension_config",
    "invalid_extension_txt",
]


@pytest.mark.parametrize(
    "base, collection_id, item_id, exp",
    [
        # Trailing / for base URI should not raise error
        (
            "http://localhost:8082",
            "test_collection",
            "test_item",
            "http://localhost:8082/test_collection/test_item",
        ),
        (
            "http://localhost:8082/",
            "test_collection",
            "test_item",
            "http://localhost:8082/test_collection/test_item",
        ),
        # No item id should not raise error
        (
            "http://localhost:8082/",
            "test_collection",
            None,
            "http://localhost:8082/test_collection",
        ),
    ],
)
def test_parse_href(base: str, collection_id: str, item_id: str | None, exp: str) -> None:
    actual = parse_href(base, collection_id, item_id)
    assert actual == exp


@pytest.mark.parametrize(
    "url, is_local",
    [
        # Href points to a STAC API endpoitn
        ("http://128.4.3.0:8082", False),
        ("https://128.4.3.0", False),
        ("https://somedrive/somepath", False),
        ("http://localhost", False),
        # Href points to a local file or directory
        ("./generated", True),
        ("generated/csv", True),
    ],
)
def test_href_is_stac_api_endpoint(url: str, is_local: bool) -> None:
    actual = href_is_stac_api_endpoint(url)
    assert actual == is_local


@pytest.mark.httpx_mock(assert_all_responses_were_requested=False)
def test_force_write_to_stac_api_given_initial_response_ok_expects_no_resend(
    httpx_mock: pytest_httpx.HTTPXMock,
) -> None:
    httpx_mock.add_response(status_code=200, method="POST")
    # Error 500 should be raise if this test fails
    httpx_mock.add_response(status_code=500, method="PUT")
    force_write_to_stac_api("http://localhost:8082", "test_collection", {})


def test_force_write_to_stac_api_given_initial_response_409_expects_send_put_response_given_put_response_ok_expects_no_raise(
    httpx_mock: pytest_httpx.HTTPXMock,
) -> None:
    httpx_mock.add_response(status_code=409, method="POST")
    httpx_mock.add_response(status_code=200, method="PUT")
    force_write_to_stac_api("http://localhost:8082", "test_collection", {})


@pytest.mark.parametrize("code", [400, 401, 403, 404, 405, 500, 501])
def test_force_write_to_stac_api_given_initial_response_4xx_or_5xx_expects_raises(
    code: int,
    httpx_mock: pytest_httpx.HTTPXMock,
) -> None:
    httpx_mock.add_response(status_code=code, method="POST")
    with pytest.raises(httpx.HTTPStatusError) as exp_ctx:
        force_write_to_stac_api("http://localhost:8082", "test_collection", {})
    assert exp_ctx.value.response.status_code == code


@pytest.mark.parametrize("code", [400, 401, 403, 404, 405, 500, 501])
def test_force_write_to_stac_api_given_initial_response_409_final_response_expects_4xx_or_5xx_raises(
    code: int,
    httpx_mock: pytest_httpx.HTTPXMock,
) -> None:
    httpx_mock.add_response(status_code=409, method="POST")
    httpx_mock.add_response(status_code=code, method="PUT")
    with pytest.raises(httpx.HTTPStatusError) as exp_ctx:
        force_write_to_stac_api("http://localhost:8082", "test_collection", {})
    assert exp_ctx.value.response.status_code == code


@pytest.mark.parametrize("file_path", INVALID_CONFIG_FILES, ids=INVALID_CONFIG_FILES_IDS)
def test_read_source_config_given_invalid_file_expects_raises(file_path: str) -> None:
    with pytest.raises(Exception):
        read_source_config(file_path)


def test_read_source_config_given_valid_csv_file_expects_correct_config_output() -> None:
    actual = read_source_config(VALID_CSV_CONFIG_FILE)
    actual[0]["column_info"] = json.loads(actual[0]["column_info"])
    assert actual == CONFIG_OUTPUT


@pytest.mark.parametrize("href", VALID_CONFIG_FILES)
def test_read_source_config_given_valid_local_files_expects_correct_config_output(
    href: str,
) -> None:
    actual = read_source_config(href)
    assert actual == CONFIG_OUTPUT


SINGLE_POINT_ITEM = pystac.Item(
    id="point_item",
    geometry=shapely.Point(150.5471916, -24.33986861),
    bbox=[150.5471916, -24.34031206, 150.5505183, -24.33986861],
    datetime=datetime.datetime(2017, 1, 1, 0, 0, 0, tzinfo=datetime.UTC),
    properties={},
)
SINGLE_POINT_ITEM_NO_START_END_DATETIME = pystac.Item(
    id="point_item",
    geometry=shapely.Point(150.5471916, -24.33986861),
    bbox=[150.5471916, -24.34031206, 150.5505183, -24.33986861],
    datetime=datetime.datetime(2017, 1, 1, 0, 0, 0, tzinfo=datetime.UTC),
    properties={},
)
MULTIPOINTS_ITEM = pystac.Item(
    id="points_item",
    geometry=shapely.MultiPoint([[150.5571916, -24.34986861], [150.5515183, -24.34131206]]),
    bbox=[150.5515183, -24.34986861, 150.5571916, -24.34131206],
    datetime=datetime.datetime(2017, 1, 2, 0, 0, 0, tzinfo=datetime.UTC),
    properties={},
)

EXP_SPATIAL_EXTENT = pystac.SpatialExtent([150.5471916, -24.34986861, 150.5571916, -24.33986861])
EXP_TEMPORAL_EXTENT = pystac.TemporalExtent(
    [
        [
            datetime.datetime(2017, 1, 1, 0, 0, 0, tzinfo=datetime.UTC),
            datetime.datetime(2017, 1, 2, 0, 0, 0, tzinfo=datetime.UTC),
        ]
    ]
)


def test_get_collection_spatial_extent() -> None:
    actual = CollectionGenerator.spatial_extent([SINGLE_POINT_ITEM, MULTIPOINTS_ITEM])
    assert actual.bboxes == EXP_SPATIAL_EXTENT.bboxes


def test_get_collection_temporal_extent() -> None:
    actual = CollectionGenerator.temporal_extent([SINGLE_POINT_ITEM, MULTIPOINTS_ITEM])
    assert actual.intervals == EXP_TEMPORAL_EXTENT.intervals


#######################################################################
# Test geometry method
#######################################################################

GEOMETRY_TEST_SET = {
    "ONE_POINT": (
        gpd.GeoDataFrame(crs="EPSG:4326", data={"geometry": [Point(1, 2)]}),
        Point(1, 2),
    ),
    "MULTIPOINT": (
        gpd.GeoDataFrame(crs="EPSG:4326", data={"geometry": [Point(1, 2), Point(3, 4)]}),
        MultiPoint(((1, 2), (3, 4))),
    ),
    "POINT_AND_MULTIPOINT": (
        gpd.GeoDataFrame(
            crs="EPSG:4326",
            data={"geometry": [Point(1, 2), Point(3, 4), MultiPoint(((5, 6), (7, 8)))]},
        ),
        MultiPoint(((1, 2), (3, 4), (5, 6), (7, 8))),
    ),
    "LINESTRING": (
        gpd.GeoDataFrame(crs="EPSG:4326", data={"geometry": [LineString(((1, 2), (3, 4)))]}),
        LineString(((1, 2), (3, 4))),
    ),
    "MULTILINESTRING": (
        gpd.GeoDataFrame(
            crs="EPSG:4326",
            data={
                "geometry": [
                    LineString(((1, 2), (3, 4))),
                    LineString(((5, 6), (7, 8))),
                ]
            },
        ),
        MultiLineString((((1, 2), (3, 4)), ((5, 6), (7, 8)))),
    ),
    "LINESTRING_AND_MULTILINESTRING": (
        gpd.GeoDataFrame(
            crs="EPSG:4326",
            data={
                "geometry": [
                    LineString(((1, 2), (3, 4))),
                    LineString(((5, 6), (7, 8))),
                    MultiLineString(
                        [
                            ((9, 10), (11, 12)),
                            ((13, 14), (15, 16)),
                        ]
                    ),
                ]
            },
        ),
        MultiLineString(
            [
                ((1, 2), (3, 4)),
                ((5, 6), (7, 8)),
                ((9, 10), (11, 12)),
                ((13, 14), (15, 16)),
            ]
        ),
    ),
    "POLYGON": (
        gpd.GeoDataFrame(
            crs="EPSG:4326",
            data={
                "geometry": [
                    Polygon(((1, 2), (3, 4), (5, 6), (1, 2))),
                ]
            },
        ),
        Polygon(((1, 2), (3, 4), (5, 6), (1, 2))),
    ),
    "MULTIPOLYGON": (
        gpd.GeoDataFrame(
            crs="EPSG:4326",
            data={
                "geometry": [
                    Polygon(((1, 2), (3, 4), (5, 6), (1, 2))),
                    Polygon(((7, 8), (9, 10), (11, 12), (7, 8))),
                ]
            },
        ),
        MultiPolygon(
            [
                (((1, 2), (3, 4), (5, 6), (1, 2)), None),
                (((7, 8), (9, 10), (11, 12), (7, 8)), None),
            ]
        ),
    ),
    "POLYGON_AND_MULTIPOLYGON": (
        gpd.GeoDataFrame(
            crs="EPSG:4326",
            data={
                "geometry": [
                    Polygon(((1, 2), (3, 4), (5, 6), (1, 2))),
                    Polygon(((7, 8), (9, 10), (11, 12), (7, 8))),
                    MultiPolygon(
                        [
                            (((13, 14), (15, 16), (17, 18), (13, 14)), None),
                            (((19, 20), (21, 22), (23, 24), (19, 20)), None),
                        ]
                    ),
                ]
            },
        ),
        MultiPolygon(
            [
                (((1, 2), (3, 4), (5, 6), (1, 2)), None),
                (((7, 8), (9, 10), (11, 12), (7, 8)), None),
                (((13, 14), (15, 16), (17, 18), (13, 14)), None),
                (((19, 20), (21, 22), (23, 24), (19, 20)), None),
            ]
        ),
    ),
    "COMPOSITE": (
        gpd.GeoDataFrame(
            crs="EPSG:4326", data={"geometry": [Point(1, 2), LineString(((3, 4), (5, 6)))]}
        ),
        Polygon(((5, 2), (5, 6), (1, 6), (1, 2), (5, 2))),
    ),
    "MORE_THAN_10": (
        gpd.GeoDataFrame(
            crs="EPSG:4326",
            data={
                "geometry": [
                    Point(1, 2),
                    Point(3, 4),
                    Point(5, 6),
                    Point(7, 8),
                    Point(9, 10),
                    Point(11, 12),
                    Point(13, 14),
                    Point(15, 16),
                    Point(17, 18),
                    Point(19, 20),
                    Point(21, 22),
                ]
            },
        ),
        Polygon(((21, 2), (21, 22), (1, 22), (1, 2), (21, 2))),
    ),
}


@pytest.mark.parametrize("df, geom", GEOMETRY_TEST_SET.values(), ids=GEOMETRY_TEST_SET.keys())
def test_geometry(df: gpd.GeoDataFrame, geom: Geometry) -> None:
    actual = VectorGenerator.geometry(df)
    assert actual == geom


def test_read_non_existent_vector_expects_throw() -> None:
    with pytest.raises(SourceAssetException):
        read_vector_asset("non_existent.geojson")


def test_vector_non_existent_layer_expects_throw() -> None:
    with pytest.raises(StacConfigException):
        read_vector_asset("tests/files/unit_tests/vectors/SA2.zip", layer="NonExistentLayer")


def test_vector_incorrectly_formatted_file_expects_throw() -> None:
    with pytest.raises(SourceAssetException):
        read_vector_asset("tests/files/unit_tests/vectors/invalid.geojson")


def test_read_non_existent_csv_expects_throw() -> None:
    with pytest.raises(SourceAssetException):
        _read_csv("non_existent.csv")


@pytest.mark.parametrize(
    "data, tzinfo, expected",
    [
        # Single naive timestamp
        (
            pd.Timestamp("2023-04-01 12:00:00"),
            "America/New_York",
            pd.Timestamp("2023-04-01 16:00:00", tz="UTC"),
        ),
        # Single tz-aware timestamp (no change)
        (
            pd.Timestamp("2023-04-01 16:00:00", tz="UTC"),
            "America/New_York",
            pd.Timestamp("2023-04-01 16:00:00", tz="UTC"),
        ),
        # Series of mixed timestamps
        (
            pd.Series(
                [
                    pd.Timestamp("2023-04-01 12:00:00"),
                    pd.Timestamp("2023-04-01 16:00:00", tz="UTC"),
                ]
            ),
            "America/New_York",
            pd.Series(
                [
                    pd.Timestamp("2023-04-01 16:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 16:00:00", tz="UTC"),
                ]
            ),
        ),
        # Series of non timestamps
        (
            pd.Series(
                [
                    pd.Timestamp("2023-04-01 12:00:00"),
                    pd.Timestamp("2023-04-01 13:00:00"),
                    pd.Timestamp("2023-04-01 14:00:00"),
                    pd.Timestamp("2023-04-01 15:00:00"),
                    pd.Timestamp("2023-04-01 20:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 17:00:00"),
                    pd.Timestamp("2023-04-01 22:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-02 09:00:00", tz="Australia/Sydney"),
                ]
            ),
            "America/New_York",
            pd.Series(
                [
                    pd.Timestamp("2023-04-01 16:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 17:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 18:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 19:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 20:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 21:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 22:00:00", tz="UTC"),
                    pd.Timestamp("2023-04-01 23:00:00", tz="UTC"),
                ]
            ),
        ),
    ],
)
def test_localise_timezone_valid(
    data: pd.Series, tzinfo: str, expected: pd.Series | pd.Timestamp
) -> None:
    result = localise_timezone(data, tzinfo)
    if isinstance(result, pd.Series):  # type: ignore[unreachable]
        pd.testing.assert_series_equal(result, expected)  # type: ignore[arg-type]
    else:
        assert result == expected  # type: ignore[unreachable]


def test_localise_timezone_invalid() -> None:
    with pytest.raises(TimezoneException):
        localise_timezone(pd.Timestamp("2020-01-01"), "Invalid")



from __future__ import annotations

from stac_generator.core.base.schema import HasColumnInfo, SourceConfig


class PointConfig(SourceConfig, HasColumnInfo):
    """Source config for point(csv) data"""

    X: str
    """Column to be treated as longitude/X coordinate"""
    Y: str
    """Column to be treated as latitude/Y coordinate"""
    Z: str | None = None
    """Column to be treated as altitude/Z coordinate"""
    T: str | None = None
    """Column to be treated as time coordinate"""
    date_format: str = "ISO8601"
    """Format to parse dates - will be used if T column is provided"""
    epsg: int = 4326
    """EPSG code"""



import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# Attempt to disable logging from other modules - does not work yet
disable_logging = ["httpcore", "httpx"]

for name in logging.root.manager.loggerDict:
    if not name.startswith(__name__) or name in disable_logging:
        logging.getLogger(name).disabled = True
